"""Declarations Abstract Syntax Tree (AST)

This file provides the definitions of various nodes used to represent a C syntax tree.
It is utilized by `pycparser` to construct and process abstract syntax trees (ASTs) for C code.
The nodes defined here implement the visitor pattern, allowing tree traversal and manipulation.

Based on the c_ast.py from the pycparser project by Eli Bendersky (https://eli.thegreenplace.net)
under the BSD license.

This file can be imported and contains the following classes:

    * Node:            Abstract base class for all AST nodes.
    * NodeVisitor:     Base class for visitor implementations for traversing AST nodes.
    * ArrayDecl:        Represents an array declaration node.
    * ArrayRef:         Represents an array reference node.
    * Assignment:       Represents an assignment operation node.
    * BinaryOp:         Represents a binary operation node.
    * Compound:         Represents a compound statement node.
    * CompoundLiteral:  Represents a compound literal node.
    * Decl:             Represents a declaration node.
    * DeclList:         Represents a list of declarations node.
    * EmptyStatement:   Represents an empty statement node.
    * ExprList:         Represents a list of expressions node.
    * FileAST:          Represents the root node of the syntax tree, containing the file's AST.
    * FuncCall:         Represents a function call node.
    * FuncDecl:         Represents a function declaration node.
    * FuncDef:          Represents a function definition node.
    * ID:               Represents an identifier node.
    * IdentifierType:   Represents a type identifier node.
    * InitList:         Represents an initialization list node.
    * NamedInitializer: Represents a named initializer node.
    * ParamList:        Represents a list of parameters node.
    * RangeDecl:        Represents a range declaration node.
    * Return:           Represents a return statement node.
    * Struct:           Represents a struct declaration node.
    * StructRef:        Represents a struct reference node.
    * TypeDecl:         Represents a type declaration node.
    * Typedef:          Represents a type definition node.
    * Typename:         Represents a type name node.
    * UnaryOp:          Represents a unary operation node.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
import logging

from uppaal2jetracer.parser.declarationparser.plyparser import Coord

logger = logging.getLogger("declarations")


class Node(ABC):
    """
    Abstract base class for AST nodes.

    Each node represents a specific construct in the C code
    and can be visited using a `NodeVisitor`.
    """

    __slots__ = ()

    def __init__(self):
        logger.debug("Created %s node: %s", self.__class__.__name__, self)

    def __repr__(self):
        """
        Generates a string representation of the node and its attributes.
        """
        attr_values = ", ".join(
            f"{slot}={getattr(self, slot, None)}" for slot in self.__slots__
        )
        return f"({self.__class__.__name__}: {attr_values})"

    @abstractmethod
    def accept(self, node_visitor: NodeVisitor):
        """
        Abstract method to accept a visitor for traversing or processing the node.

        :param node_visitor: NodeVisitor, which implements
        the specific visit method for this type of node.
        :type node_visitor: NodeVisitor
        :return: Depends on the NodeVisitor
        """

    @abstractmethod
    def children(self):
        """
        Returns a sequence of all child nodes.

        :return: An iterable containing the child nodes of the current node.
        """

class NodeVisitor(ABC):
    """
    Base class for visiting AST nodes using the Visitor pattern.

    This class defines a set of methods, one for each type of AST node,
    which must be implemented by subclasses.
    """

    #_method_cache = None

    # def visit(self, node):
    #     """ Visit a node.
    #     """
    #
    #     if self._method_cache is None:
    #         self._method_cache = {}
    #
    #     visitor = self._method_cache.get(node.__class__.__name__, None)
    #     if visitor is None:
    #         method = 'visit_' + node.__class__.__name__
    #         visitor = getattr(self, method, self.generic_visit)
    #         self._method_cache[node.__class__.__name__] = visitor
    #
    #     return visitor(node)
    #
    # def generic_visit(self, node):
    #     """ Called if no explicit visitor function exists for a
    #         node. Implements preorder visiting of the node.
    #     """
    #     for c in node:
    #         self.visit(c)

    @abstractmethod
    def visit_arraydecl(self, array_decl: ArrayDecl):
        """
        Abstract method to visit an :class:`ArrayDecl` node.
        Return depends on implementation of this method.

        :param array_decl: The array declaration node.
        :type array_decl: ArrayDecl
        """

    @abstractmethod
    def visit_arrayref(self, array_ref: ArrayRef):
        """
        Abstract method to visit an :class:`ArrayRef` node.
        Return depends on implementation of this method.

        :param array_ref: The array reference node.
        :type array_ref: ArrayRef
        """

    @abstractmethod
    def visit_assignment(self, assignment: Assignment):
        """
        Abstract method to visit an :class:`Assignment` node.
        Return depends on implementation of this method.

        :param assignment: The assignment node.
        :type assignment: Assignment
        """

    @abstractmethod
    def visit_binaryop(self, binaryop: BinaryOp):
        """
        Abstract method to visit a :class:`BinaryOp` node.
        Return depends on implementation of this method.

        :param binaryop: The binary operation node.
        :type binaryop: BinaryOp
        """

    @abstractmethod
    def visit_compound(self, compound: Compound):
        """
        Abstract method to visit a :class:`Compound` node.
        Return depends on implementation of this method.

        :param compound: The compound node.
        :type compound: Compound
        """

    @abstractmethod
    def visit_compoundliteral(self, compound_literal: CompoundLiteral):
        """
        Abstract method to visit a :class:`CompoundLiteral` node.
        Return depends on implementation of this method.

        :param compound_literal: The compound literal node.
        :type compound_literal: CompoundLiteral
        """

    @abstractmethod
    def visit_constant(self, constant: Constant):
        """
        Abstract method to visit a :class:`Constant` node.
        Return depends on implementation of this method.

        :param constant: The constant node.
        :type constant: Constant
        """

    @abstractmethod
    def visit_decl(self, decl: Decl):
        """
        Abstract method to visit a :class:`Decl` node.
        Return depends on implementation of this method.

        :param decl: The declaration node.
        :type decl: Decl
        """

    @abstractmethod
    def visit_decllist(self, decl_list: DeclList):
        """
        Abstract method to visit a :class:`DeclList` node.
        Return depends on implementation of this method.

        :param decl_list: The declaration list node.
        :type decl_list: DeclList
        """

    @abstractmethod
    def visit_emptystatement(self, empty_statement: EmptyStatement):
        """
        Abstract method to visit an :class:`EmptyStatement` node.
        Return depends on implementation of this method.

        :param empty_statement: The empty statement node.
        :type empty_statement: EmptyStatement
        """

    @abstractmethod
    def visit_exprlist(self, expr_list: ExprList):
        """
        Abstract method to visit an :class:`ExprList` node.
        Return depends on implementation of this method.

        :param expr_list: The expression list node.
        :type expr_list: ExprList
        """

    @abstractmethod
    def visit_fileast(self, fileast: FileAST):
        """
        Abstract method to visit a :class:`FileAST` node.
        Return depends on implementation of this method.

        :param fileast: The fileast node.
        :type fileast: FileASt
        """

    @abstractmethod
    def visit_funccall(self, funccall: FuncCall):
        """
        Abstract method to visit a :class:`FuncCall` node.
        Return depends on implementation of this method.

        :param funccall: The function call node.
        :type funccall: FuncCall
        """

    @abstractmethod
    def visit_funcdecl(self, funcdecl: FuncDecl):
        """
        Abstract method to visit a :class:`FuncDecl` node.
        Return depends on implementation of this method.

        :param funcdecl: The function declaration node.
        :type funcdecl: FuncDecl
        """

    @abstractmethod
    def visit_funcdef(self, funcdef: FuncDef):
        """
        Abstract method to visit a :class:`FuncDef` node.
        Return depends on implementation of this method.

        :param funcdef: The function definition node.
        :type funcdef: FuncDef
        """

    @abstractmethod
    def visit_id(self, identifier: ID):
        """
        Abstract method to visit an :class:`ID` node.
        Return depends on implementation of this method.

        :param identifier: The identifier node.
        :type identifier: ID
        """

    @abstractmethod
    def visit_identifiertype(self, identifier_type: IdentifierType):
        """
        Abstract method to visit an :class:`IdentifierType` node.
        Return depends on implementation of this method.

        :param identifier_type: The identifier type node.
        :type identifier_type: IdentifierType
        """

    @abstractmethod
    def visit_initlist(self, init_list: InitList):
        """
        Abstract method to visit an :class:`InitList` node.
        Return depends on implementation of this method.

        :param init_list: The initialization list node.
        :type init_list: InitList
        """

    @abstractmethod
    def visit_namedinitializer(self, named_initializer: NamedInitializer):
        """
        Abstract method to visit a :class:`NamedInitializer` node.
        Return depends on implementation of this method.

        :param named_initializer: The named initializer node.
        :type named_initializer: NamedInitializer
        """

    @abstractmethod
    def visit_paramlist(self, param_list: ParamList):
        """
        Abstract method to visit a :class:`ParamList` node.
        Return depends on implementation of this method.

        :param param_list: The parameter list node.
        :type param_list: ParamList
        """

    @abstractmethod
    def visit_rangedecl(self, range_decl: RangeDecl):
        """
        Abstract method to visit a :class:`RangeDecl` node.
        Return depends on implementation of this method.

        :param range_decl: The range declaration node.
        :type range_decl: RangeDecl
        """

    @abstractmethod
    def visit_return(self, ret: Return):
        """
        Abstract method to visit a :class:`Return` node.
        Return depends on implementation of this method.

        :param ret: The return node.
        :type ret: Return
        """

    @abstractmethod
    def visit_struct(self, struct: Struct):
        """
        Abstract method to visit a :class:`Struct` node.
        Return depends on implementation of this method.

        :param struct: The struct node.
        :type struct: Struct
        """

    @abstractmethod
    def visit_structref(self, structref: StructRef):
        """
        Abstract method to visit a :class:`StructRef` node.
        Return depends on implementation of this method.

        :param structref: The struct reference node.
        :type structref: StructRef
        """

    @abstractmethod
    def visit_typedecl(self, typedecl: TypeDecl):
        """
        Abstract method to visit a :class:`TypeDecl` node.
        Return depends on implementation of this method.

        :param typedecl: The type declaration node.
        :type typedecl: TypeDecl
        """

    @abstractmethod
    def visit_typedef(self, typedef: Typedef):
        """
        Abstract method to visit a :class:`Typedef` node.
        Return depends on implementation of this method.

        :param typedef: The type definition node.
        :type typedef: Typedef
        """

    @abstractmethod
    def visit_typename(self, typename: Typename):
        """
        Abstract method to visit a :class:`Typename` node.
        Return depends on implementation of this method.

        :param typename: The type name node.
        :type typename: Typename
        """

    @abstractmethod
    def visit_unaryop(self, unaryop: UnaryOp):
        """
        Abstract method to visit a :class:`UnaryOp` node.
        Return depends on implementation of this method.

        :param unaryop: The unary operation node.
        :type unaryop: UnaryOp
        """


class ArrayDecl(Node):
    """
    Represents a nested declaration of an array with the given type.

    :ivar type: The base type of the array.
    :vartype type: Node
    :ivar dim: The dimension (e.g., constant 42).
    :vartype dim: int
    :ivar dim_quals: List of dimension qualifiers,
    supporting C99's 'const' and 'static' in array dimensions.
    :vartype dim_quals: list
    """

    __slots__ = ("type", "dim", "dim_quals", "coord", "__weakref__")

    def __init__(self, type, dim, dim_quals, coord = None):
        self.type = type
        self.dim = dim
        self.dim_quals = dim_quals
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.type is not None:
            nodelist.append(("type", self.type))
        if self.dim is not None:
            nodelist.append(("dim", self.dim))
        return tuple(nodelist)

    def __iter__(self):
        if self.type is not None:
            yield self.type
        if self.dim is not None:
            yield self.dim

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_arraydecl(self)

    attr_names = ('dim_quals',)

class ArrayRef(Node):
    """
    Represents an array reference in the AST.

    :ivar name: The name of the array being referenced.
    :vartype name: Node
    :ivar subscript: The index or subscript of the array.
    :vartype subscript: Node
    """

    __slots__ = ('name', 'subscript', 'coord', '__weakref__')

    def __init__(self, name, subscript, coord = None):
        self.name = name
        self.subscript = subscript
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.name is not None:
            nodelist.append(("name", self.name))
        if self.subscript is not None:
            nodelist.append(("subscript", self.subscript))
        return tuple(nodelist)

    def __iter__(self):
        if self.name is not None:
            yield self.name
        if self.subscript is not None:
            yield self.subscript

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_arrayref(self)

    attr_names = ()

class Assignment(Node):
    """
    Represents an assignment operation in the AST.

    :ivar op: The assignment operator (e.g., "=", "+=").
    :vartype op: str
    :ivar lvalue: The left-hand side of the assignment.
    :vartype lvalue: Node
    :ivar rvalue: The right-hand side of the assignment.
    :vartype rvalue: Node
    """

    __slots__ = ("op", "lvalue", "rvalue", "coord", "__weakref__")
    __annotations__ = {
        "op": str,
        "lvalue": Node,
        "rvalue": Node,
        "coord": Coord,
    }

    def __init__(self, op, lvalue, rvalue, coord = None):
        self.op = op
        self.lvalue = lvalue
        self.rvalue = rvalue
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.lvalue is not None:
            nodelist.append(("lvalue", self.lvalue))
        if self.rvalue is not None:
            nodelist.append(("rvalue", self.rvalue))
        return tuple(nodelist)

    def __iter__(self):
        if self.lvalue is not None:
            yield self.lvalue
        if self.rvalue is not None:
            yield self.rvalue

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_assignment(self)

    attr_names = ("op",)

#
class Alignas(Node):
    __slots__ = ('alignment', 'coord', '__weakref__')

    def __init__(self, alignment, coord = None):
        self.alignment = alignment
        self.coord = coord

    def children(self):
        nodelist = []
        if self.alignment is not None:
            nodelist.append(("alignment", self.alignment))
        return tuple(nodelist)

    def __iter__(self):
        if self.alignment is not None:
            yield self.alignment

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

class BinaryOp(Node):
    """
    Represents a binary operation in the AST.

    :ivar op: The binary operator (e.g., "+", "-", "*", "/").
    :vartype op: str
    :ivar left: The left operand of the binary operation.
    :vartype left: Node
    :ivar right: The right operand of the binary operation.
    :vartype right: Node
    """

    __slots__ = ("op", "left", "right", "coord", "__weakref__")

    def __init__(self, op, left, right, coord = None):
        self.op = op
        self.left = left
        self.right = right
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.left is not None:
            nodelist.append(("left", self.left))
        if self.right is not None:
            nodelist.append(("right", self.right))
        return tuple(nodelist)

    def __iter__(self):
        if self.left is not None:
            yield self.left
        if self.right is not None:
            yield self.right

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_binaryop(self)

    attr_names = ("op",)

#
class Break(Node):
    __slots__ = ('coord', '__weakref__')

    def __init__(self, coord = None):
        self.coord = coord

    def children(self):
        return ()

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

#
class Case(Node):
    __slots__ = ("expr", "stmts", "coord", "__weakref__")

    def __init__(self, expr, stmts, coord = None):
        self.expr = expr
        self.stmts = stmts
        self.coord = coord

    def children(self):
        nodelist = []
        if self.expr is not None:
            nodelist.append(("expr", self.expr))
        for i, child in enumerate(self.stmts or []):
            nodelist.append((f"stmts[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        if self.expr is not None:
            yield self.expr
        yield from (self.stmts or [])

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

#
class Cast(Node):
    __slots__ = ("to_type", "expr", "coord", "__weakref__")

    def __init__(self, to_type, expr, coord = None):
        self.to_type = to_type
        self.expr = expr
        self.coord = coord

    def children(self):
        nodelist = []
        if self.to_type is not None:
            nodelist.append(("to_type", self.to_type))
        if self.expr is not None:
            nodelist.append(("expr", self.expr))
        return tuple(nodelist)

    def __iter__(self):
        if self.to_type is not None:
            yield self.to_type
        if self.expr is not None:
            yield self.expr

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

class Compound(Node):
    """
    Represents a compound statement (block) in the AST.

    :ivar block_items: A list of statements or declarations within the block.
    :vartype block_items: list[Node]
    """

    __slots__ = ("block_items", "coord", "__weakref__")

    def __init__(self, block_items, coord = None):
        self.block_items = block_items
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        for i, child in enumerate(self.block_items or []):
            nodelist.append((f"block_items[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        yield from (self.block_items or [])

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_compound(self)

    attr_names = ()

class CompoundLiteral(Node):
    """
    Represents a compound literal (anonymous aggregate).

    :ivar type: The type of the compound literal.
    :vartype type: Node
    :ivar init: The initializer for the compound literal.
    :vartype init: Node
    """

    __slots__ = ("type", "init", "coord", "__weakref__")

    def __init__(self, type, init, coord = None):
        self.type = type
        self.init = init
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.type is not None:
            nodelist.append(("type", self.type))
        if self.init is not None:
            nodelist.append(("init", self.init))
        return tuple(nodelist)

    def __iter__(self):
        if self.type is not None:
            yield self.type
        if self.init is not None:
            yield self.init

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_compoundliteral(self)

    attr_names = ()

class Constant(Node):
    """
    Represents a constant value in the AST.

    :ivar type: The type of the constant (e.g., "int", "double", "char", "string").
    :vartype type: str
    :ivar value: The value of the constant.
    :vartype value: str
    """

    __slots__ = ("type", "value", "coord", "__weakref__")

    def __init__(self, type, value, coord = None):
        self.type = type
        self.value = value
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        return tuple(nodelist)

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_constant(self)

    attr_names = ('type', 'value',)

#
class Continue(Node):
    __slots__ = ("coord", "__weakref__")

    def __init__(self, coord = None):
        self.coord = coord

    def children(self):
        return ()

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

class Decl(Node):
    """
    Represents a variable or type declaration in the AST.

    :ivar name: The name of the declared variable or type.
    :vartype name: str
    :ivar quals: A list of qualifiers (e.g., "const", "volatile").
    :vartype quals: list[str]
    :ivar type: The declaration type, possibly nested with modifiers.
    :vartype type: Node
    :ivar init: The initialization value, or None if not initialized.
    :vartype init: Node
    """

    __slots__ = ("name", "quals", "align", "storage", "funcspec",
                 "type", "init", "bitsize", "coord", "__weakref__")

    def __init__(self, name, quals, align, storage, funcspec,
                 type, init, bitsize, coord = None):
        self.name = name
        self.quals = quals
        self.align = align
        self.storage = storage
        self.funcspec = funcspec
        self.type = type
        self.init = init
        self.bitsize = bitsize
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.type is not None:
            nodelist.append(("type", self.type))
        if self.init is not None:
            nodelist.append(("init", self.init))
        if self.bitsize is not None:
            nodelist.append(("bitsize", self.bitsize))
        return tuple(nodelist)

    def __iter__(self):
        if self.type is not None:
            yield self.type
        if self.init is not None:
            yield self.init
        if self.bitsize is not None:
            yield self.bitsize

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_decl(self)

    attr_names = ("name", "quals", "align", "storage", "funcspec",)

class DeclList(Node):
    """
    Represents a list of declarations in the AST.

    :ivar decls: A list of declarations.
    :vartype decls: list[Node]
    """

    __slots__ = ("decls", "coord", "__weakref__")

    def __init__(self, decls, coord = None):
        self.decls = decls
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        for i, child in enumerate(self.decls or []):
            nodelist.append((f"decls[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        yield from (self.decls or [])

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_decllist(self)

    attr_names = ()

#
class Default(Node):
    __slots__ = ("stmts", "coord", "__weakref__")

    def __init__(self, stmts, coord = None):
        self.stmts = stmts
        self.coord = coord

    def children(self):
        nodelist = []
        for i, child in enumerate(self.stmts or []):
            nodelist.append((f"stmts[{i}]" % i, child))
        return tuple(nodelist)

    def __iter__(self):
        yield from (self.stmts or [])

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

#
class DoWhile(Node):
    __slots__ = ("cond", "stmt", "coord", "__weakref__")

    def __init__(self, cond, stmt, coord = None):
        self.cond = cond
        self.stmt = stmt
        self.coord = coord

    def children(self):
        nodelist = []
        if self.cond is not None:
            nodelist.append(("cond", self.cond))
        if self.stmt is not None:
            nodelist.append(("stmt", self.stmt))
        return tuple(nodelist)

    def __iter__(self):
        if self.cond is not None:
            yield self.cond
        if self.stmt is not None:
            yield self.stmt

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

#
class EllipsisParam(Node):
    __slots__ = ("coord", "__weakref__")

    def __init__(self, coord = None):
        self.coord = coord

    def children(self):
        return ()

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

class EmptyStatement(Node):
    """
    Represents an empty statement (a semicolon ";" on its own).
    """

    __slots__ = ("coord", "__weakref__")

    def __init__(self, coord = None):
        self.coord = coord
        super().__init__()

    def children(self):
        return ()

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_emptystatement(self)

    attr_names = ()

#
class Enum(Node):
    __slots__ = ("name", "values", "coord", "__weakref__")

    def __init__(self, name, values, coord = None):
        self.name = name
        self.values = values
        self.coord = coord

    def children(self):
        nodelist = []
        if self.values is not None:
            nodelist.append(("values", self.values))
        return tuple(nodelist)

    def __iter__(self):
        if self.values is not None:
            yield self.values

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ('name',)

#
class Enumerator(Node):
    __slots__ = ("name", "value", "coord", "__weakref__")

    def __init__(self, name, value, coord = None):
        self.name = name
        self.value = value
        self.coord = coord

    def children(self):
        nodelist = []
        if self.value is not None:
            nodelist.append(("value", self.value))
        return tuple(nodelist)

    def __iter__(self):
        if self.value is not None:
            yield self.value

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ('name',)

#
class EnumeratorList(Node):
    __slots__ = ('enumerators', 'coord', '__weakref__')

    def __init__(self, enumerators, coord = None):
        self.enumerators = enumerators
        self.coord = coord

    def children(self):
        nodelist = []
        for i, child in enumerate(self.enumerators or []):
            nodelist.append((f"enumerators[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        for child in (self.enumerators or []):
            yield child

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

class ExprList(Node):
    """
    Represents a list of expressions separated by the comma operator.

    :ivar exprs: A list of expressions.
    :vartype exprs: list[Node]
    """

    __slots__ = ("exprs", "coord", "__weakref__")

    def __init__(self, exprs, coord = None):
        self.exprs = exprs
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        for i, child in enumerate(self.exprs or []):
            nodelist.append((f"exprs[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        yield from (self.exprs or [])

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_exprlist(self)

    attr_names = ()

class FileAST(Node):
    """
    Represents the top-level AST node for an UPPAAL file (a translation unit).

    :ivar ext: A list of external declarations or definitions in the file.
    :vartype ext: list[Node]
    """

    __slots__ = ("ext", "coord", "__weakref__")
    __annotations__ = {}

    def __init__(self, ext, coord = None):
        self.ext = ext
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        for i, child in enumerate(self.ext or []):
            nodelist.append((f"ext[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        yield from (self.ext or [])

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_fileast(self)

    attr_names = ()

#
class For(Node):
    __slots__ = ("init", "cond", "next", "stmt", "coord", "__weakref__")

    def __init__(self, init, cond, next_, stmt, coord = None):
        self.init = init
        self.cond = cond
        self.next = next_
        self.stmt = stmt
        self.coord = coord

    def children(self):
        nodelist = []
        if self.init is not None:
            nodelist.append(("init", self.init))
        if self.cond is not None:
            nodelist.append(("cond", self.cond))
        if self.next is not None:
            nodelist.append(("next", self.next))
        if self.stmt is not None:
            nodelist.append(("stmt", self.stmt))
        return tuple(nodelist)

    def __iter__(self):
        if self.init is not None:
            yield self.init
        if self.cond is not None:
            yield self.cond
        if self.next is not None:
            yield self.next
        if self.stmt is not None:
            yield self.stmt

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

class FuncCall(Node):
    """
    Represents a function call in the AST.

    :ivar name: The name of the function being called.
    :vartype name: str
    :ivar args: The arguments passed to the function.
    :vartype args: Node
    """

    __slots__ = ("name", "args", "coord", "__weakref__")

    def __init__(self, name, args, coord = None):
        self.name = name
        self.args = args
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.name is not None:
            nodelist.append(("name", self.name))
        if self.args is not None:
            nodelist.append(("args", self.args))
        return tuple(nodelist)

    def __iter__(self):
        if self.name is not None:
            yield self.name
        if self.args is not None:
            yield self.args

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_funccall(self)

    attr_names = ()

class FuncDecl(Node):
    """
    Represents a function declaration in the AST.

    :ivar args: The parameters of the function.
    :vartype args: Node
    :ivar type: The return type of the function.
    :vartype type: Node
    """

    __slots__ = ("args", "type", "coord", "__weakref__")

    def __init__(self, args, type, coord = None):
        self.args = args
        self.type = type
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.args is not None:
            nodelist.append(("args", self.args))
        if self.type is not None:
            nodelist.append(("type", self.type))
        return tuple(nodelist)

    def __iter__(self):
        if self.args is not None:
            yield self.args
        if self.type is not None:
            yield self.type

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_funcdecl(self)

    attr_names = ()

class FuncDef(Node):
    """
    Represents a function definition in the AST.

    :ivar decl: The declarator for the function name.
    :vartype decl: Node
    :ivar param_decls: A list of parameter declarations (optional, for K&R-style definitions).
    :vartype param_decls: list[Node]
    :ivar body: The body of the function (a compound statement).
    :vartype body: Node
    """

    __slots__ = ("decl", "param_decls", "body", "coord", "__weakref__")

    def __init__(self, decl, param_decls, body, coord = None):
        self.decl = decl
        self.param_decls = param_decls
        self.body = body
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.decl is not None:
            nodelist.append(("decl", self.decl))
        if self.body is not None:
            nodelist.append(("body", self.body))
        for i, child in enumerate(self.param_decls or []):
            nodelist.append((f"param_decls[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        if self.decl is not None:
            yield self.decl
        if self.body is not None:
            yield self.body
        for child in (self.param_decls or []):
            yield child

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_funcdef(self)

    attr_names = ()

#
class Goto(Node):
    __slots__ = ('name', 'coord', '__weakref__')

    def __init__(self, name, coord = None):
        self.name = name
        self.coord = coord

    def children(self):
        nodelist = []
        return tuple(nodelist)

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ('name',)

class ID(Node):
    """
    Represents an identifier in the AST.

    :ivar name: The name of the identifier.
    :vartype name: str
    """

    __slots__ = ("name", "coord", "__weakref__")

    def __init__(self, name, coord = None):
        self.name = name
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        return tuple(nodelist)

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_id(self)

    attr_names = ("name",)

class IdentifierType(Node):
    """
    Represents a holder for types that are a simple identifier (e.g., built-in types like "int").

    :ivar names: A list of type names (e.g., ["unsigned", "int"]).
    :vartype names: list[str]
    """

    __slots__ = ("names", "coord", "__weakref__")

    def __init__(self, names, coord = None):
        self.names = names
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        return tuple(nodelist)

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_identifiertype(self)

    attr_names = ("names",)

#
class If(Node):
    __slots__ = ('cond', 'iftrue', 'iffalse', 'coord', '__weakref__')

    def __init__(self, cond, iftrue, iffalse, coord = None):
        self.cond = cond
        self.iftrue = iftrue
        self.iffalse = iffalse
        self.coord = coord

    def children(self):
        nodelist = []
        if self.cond is not None:
            nodelist.append(("cond", self.cond))
        if self.iftrue is not None:
            nodelist.append(("iftrue", self.iftrue))
        if self.iffalse is not None:
            nodelist.append(("iffalse", self.iffalse))
        return tuple(nodelist)

    def __iter__(self):
        if self.cond is not None:
            yield self.cond
        if self.iftrue is not None:
            yield self.iftrue
        if self.iffalse is not None:
            yield self.iffalse

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

class InitList(Node):
    """
    Represents an initialization list used for compound literals.

    :ivar exprs: A list of expressions for initialization.
    :vartype exprs: list[Node]
    """

    __slots__ = ("exprs", "coord", "__weakref__")

    def __init__(self, exprs, coord = None):
        self.exprs = exprs
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        for i, child in enumerate(self.exprs or []):
            nodelist.append((f"exprs[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        yield from (self.exprs or [])

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_initlist(self)

    attr_names = ()

#
class Label(Node):
    __slots__ = ("name", "stmt", "coord", "__weakref__")

    def __init__(self, name, stmt, coord = None):
        self.name = name
        self.stmt = stmt
        self.coord = coord

    def children(self):
        nodelist = []
        if self.stmt is not None:
            nodelist.append(("stmt", self.stmt))
        return tuple(nodelist)

    def __iter__(self):
        if self.stmt is not None:
            yield self.stmt

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ('name',)

class NamedInitializer(Node):
    """
    Represents a named initializer for C99.

    :ivar name: A sequence of nodes representing the name
    (e.g., hierarchical or constant expressions).
    :vartype name: list[Node]
    :ivar expr: The expression initializing the name.
    :vartype expr: Node
    """

    __slots__ = ("name", "expr", "coord", "__weakref__")

    def __init__(self, name, expr, coord = None):
        self.name = name
        self.expr = expr
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.expr is not None:
            nodelist.append(("expr", self.expr))
        for i, child in enumerate(self.name or []):
            nodelist.append((f"name[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        if self.expr is not None:
            yield self.expr
        yield from (self.name or [])

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_namedinitializer(self)

    attr_names = ()

class ParamList(Node):
    """
    Represents a list of function parameter declarations.

    :ivar params: A list of parameter declarations.
    :vartype params: list[Node]
    """

    __slots__ = ("params", "coord", "__weakref__")

    def __init__(self, params, coord = None):
        self.params = params
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        for i, child in enumerate(self.params or []):
            nodelist.append((f"params[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        yield from (self.params or [])

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_paramlist(self)

    attr_names = ()

#
class PtrDecl(Node):
    __slots__ = ('quals', 'type', 'coord', '__weakref__')

    def __init__(self, quals, type, coord = None):
        self.quals = quals
        self.type = type
        self.coord = coord

    def children(self):
        nodelist = []
        if self.type is not None:
            nodelist.append(("type", self.type))
        return tuple(nodelist)

    def __iter__(self):
        if self.type is not None:
            yield self.type

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ('quals',)

class RangeDecl(Node):
    __slots__ = ("type", "lower", "upper", "coord", "__weakref__")

    def __init__(self, type, lower, upper, coord = None):
        self.type = type
        self.lower = lower
        self.upper = upper
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.type is not None:
            nodelist.append(("type", self.type))
        if self.lower is not None:
            nodelist.append(("lower", self.lower))
        if self.upper is not None:
            nodelist.append(("upper", self.upper))
        return tuple(nodelist)

    def __iter__(self):
        if self.type is not None:
            yield self.type

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_rangedecl(self)

    attr_names = ()

class Return(Node):
    """
    Represents a return statement in the AST.

    :ivar expr: The expression being returned, or None if it's a void return.
    :vartype expr: Node or None
    """

    __slots__ = ("expr", "coord", "__weakref__")

    def __init__(self, expr, coord = None):
        self.expr = expr
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.expr is not None:
            nodelist.append(("expr", self.expr))
        return tuple(nodelist)

    def __iter__(self):
        if self.expr is not None:
            yield self.expr

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_return(self)

    attr_names = ()

#
class StaticAssert(Node):
    __slots__ = ('cond', 'message', 'coord', '__weakref__')

    def __init__(self, cond, message, coord = None):
        self.cond = cond
        self.message = message
        self.coord = coord

    def children(self):
        nodelist = []
        if self.cond is not None:
            nodelist.append(("cond", self.cond))
        if self.message is not None:
            nodelist.append(("message", self.message))
        return tuple(nodelist)

    def __iter__(self):
        if self.cond is not None:
            yield self.cond
        if self.message is not None:
            yield self.message

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

class Struct(Node):
    """
    Represents a struct declaration in the AST.

    :ivar name: The tag name of the struct.
    :vartype name: str
    :ivar decls: The declarations of members in the struct.
    :vartype decls: list[Node]
    """

    __slots__ = ("name", "decls", "coord", "__weakref__")

    def __init__(self, name, decls, coord = None):
        self.name = name
        self.decls = decls
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        for i, child in enumerate(self.decls or []):
            nodelist.append((f"decls[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        yield from (self.decls or [])

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_struct(self)

    attr_names = ("name",)

class StructRef(Node):
    """
    Represents a reference to a struct field.

    :ivar name: The struct being referenced.
    :vartype name: Node
    :ivar type: The type of reference ("." for member access or '->' for pointer dereference).
    :vartype type: str
    :ivar field: The field being accessed in the struct.
    :vartype field: Node
    """

    __slots__ = ("name", "type", "field", "coord", "__weakref__")

    def __init__(self, name, type, field, coord = None):
        self.name = name
        self.type = type
        self.field = field
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.name is not None:
            nodelist.append(("name", self.name))
        if self.field is not None:
            nodelist.append(("field", self.field))
        return tuple(nodelist)

    def __iter__(self):
        if self.name is not None:
            yield self.name
        if self.field is not None:
            yield self.field

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_structref(self)

    attr_names = ("type",)

#
class Switch(Node):
    __slots__ = ('cond', 'stmt', 'coord', '__weakref__')

    def __init__(self, cond, stmt, coord = None):
        self.cond = cond
        self.stmt = stmt
        self.coord = coord

    def children(self):
        nodelist = []
        if self.cond is not None:
            nodelist.append(("cond", self.cond))
        if self.stmt is not None:
            nodelist.append(("stmt", self.stmt))
        return tuple(nodelist)

    def __iter__(self):
        if self.cond is not None:
            yield self.cond
        if self.stmt is not None:
            yield self.stmt

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

#
class TernaryOp(Node):
    __slots__ = ('cond', 'iftrue', 'iffalse', 'coord', '__weakref__')

    def __init__(self, cond, iftrue, iffalse, coord = None):
        self.cond = cond
        self.iftrue = iftrue
        self.iffalse = iffalse
        self.coord = coord

    def children(self):
        nodelist = []
        if self.cond is not None:
            nodelist.append(("cond", self.cond))
        if self.iftrue is not None:
            nodelist.append(("iftrue", self.iftrue))
        if self.iffalse is not None:
            nodelist.append(("iffalse", self.iffalse))
        return tuple(nodelist)

    def __iter__(self):
        if self.cond is not None:
            yield self.cond
        if self.iftrue is not None:
            yield self.iftrue
        if self.iffalse is not None:
            yield self.iffalse

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

class TypeDecl(Node):
    """
    Represents a base type declaration in the AST.

    :ivar declname: The name of the declaration.
    :vartype declname: str
    :ivar quals: A list of qualifiers (e.g., "const", "volatile").
    :vartype quals: list[str]
    :ivar type: The type of the declaration.
    :vartype type: Node
    """

    __slots__ = ("declname", "quals", "align", "type", "coord", "__weakref__")

    def __init__(self, declname, quals, align, type, coord = None):
        self.declname = declname
        self.quals = quals
        self.align = align
        self.type = type
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.type is not None:
            nodelist.append(("type", self.type))
        return tuple(nodelist)

    def __iter__(self):
        if self.type is not None:
            yield self.type

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_typedecl(self)

    attr_names = ("declname", "quals", "align",)

class Typedef(Node):
    """
    Represents a typedef declaration in the AST.

    :ivar name: The name of the typedef.
    :vartype name: str
    :ivar quals: A list of qualifiers (e.g., "const", "volatile").
    :vartype quals: list[str]
    :ivar storage: The storage class specifier for the typedef (e.g., "static").
    :vartype storage: str
    :ivar type: The type being defined by the typedef.
    :vartype type: Node
    """

    __slots__ = ("name", "quals", "storage", "type", "coord", "__weakref__")

    def __init__(self, name, quals, storage, type, coord = None):
        self.name = name
        self.quals = quals
        self.storage = storage
        self.type = type
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.type is not None:
            nodelist.append(("type", self.type))
        return tuple(nodelist)

    def __iter__(self):
        if self.type is not None:
            yield self.type

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_typedef(self)

    attr_names = ("name", "quals", "storage",)

class Typename(Node):
    """
    Represents a typename in the AST.

    :ivar name: The name of the typename (or None for unnamed types).
    :vartype name: str or None
    :ivar quals: A list of qualifiers (e.g., "const", "volatile").
    :vartype quals: list[str]
    :ivar align: The alignment specifier for the type (if any).
    :vartype align: str or None
    :ivar type: The type being referred to by the typename.
    :vartype type: Node
    """

    __slots__ = ("name", "quals", "align", "type", "coord", "__weakref__")

    def __init__(self, name, quals, align, type, coord = None):
        self.name = name
        self.quals = quals
        self.align = align
        self.type = type
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.type is not None:
            nodelist.append(("type", self.type))
        return tuple(nodelist)

    def __iter__(self):
        if self.type is not None:
            yield self.type

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_typename(self)

    attr_names = ("name", "quals", "align",)

class UnaryOp(Node):
    """
    Represents a unary operation in the AST.

    :ivar op: The unary operator (e.g., "-", "!", "~").
    :vartype op: str
    :ivar expr: The operand of the unary operation.
    :vartype expr: Node
    """

    __slots__ = ("op", "expr", "coord", "__weakref__")

    def __init__(self, op, expr, coord = None):
        self.op = op
        self.expr = expr
        self.coord = coord
        super().__init__()

    def children(self):
        nodelist = []
        if self.expr is not None:
            nodelist.append(("expr", self.expr))
        return tuple(nodelist)

    def __iter__(self):
        if self.expr is not None:
            yield self.expr

    def accept(self, node_visitor: NodeVisitor):
        return node_visitor.visit_unaryop(self)

    attr_names = ("op",)

#
class Union(Node):
    __slots__ = ('name', 'decls', 'coord', '__weakref__')

    def __init__(self, name, decls, coord = None):
        self.name = name
        self.decls = decls
        self.coord = coord

    def children(self):
        nodelist = []
        for i, child in enumerate(self.decls or []):
            nodelist.append((f"decls[{i}]", child))
        return tuple(nodelist)

    def __iter__(self):
        yield from (self.decls or [])

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ('name',)

#
class While(Node):
    __slots__ = ('cond', 'stmt', 'coord', '__weakref__')

    def __init__(self, cond, stmt, coord = None):
        self.cond = cond
        self.stmt = stmt
        self.coord = coord

    def children(self):
        nodelist = []
        if self.cond is not None:
            nodelist.append(("cond", self.cond))
        if self.stmt is not None:
            nodelist.append(("stmt", self.stmt))
        return tuple(nodelist)

    def __iter__(self):
        if self.cond is not None:
            yield self.cond
        if self.stmt is not None:
            yield self.stmt

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ()

#
class Pragma(Node):
    __slots__ = ('string', 'coord', '__weakref__')

    def __init__(self, string, coord = None):
        self.string = string
        self.coord = coord

    def children(self):
        nodelist = []
        return tuple(nodelist)

    def accept(self, node_visitor: NodeVisitor):
        pass

    attr_names = ('string',)

#---------

def fix_switch_cases(switch_node):
    """ The 'case' statements in a 'switch' come out of parsing with one
        child node, so subsequent statements are just tucked to the parent
        Compound. Additionally, consecutive (fall-through) case statements
        come out messy. This is a peculiarity of the C grammar. The following:

            switch (myvar) {
                case 10:
                    k = 10;
                    p = k + 1;
                    return 10;
                case 20:
                case 30:
                    return 20;
                default:
                    break;
            }

        Creates this tree (pseudo-dump):

            Switch
                ID: myvar
                Compound:
                    Case 10:
                        k = 10
                    p = k + 1
                    return 10
                    Case 20:
                        Case 30:
                            return 20
                    Default:
                        break

        The goal of this transform is to fix this mess, turning it into the
        following:

            Switch
                ID: myvar
                Compound:
                    Case 10:
                        k = 10
                        p = k + 1
                        return 10
                    Case 20:
                    Case 30:
                        return 20
                    Default:
                        break

        A fixed AST node is returned. The argument may be modified.
    """
    assert isinstance(switch_node, Switch)
    if not isinstance(switch_node.stmt, Compound):
        return switch_node

    # The new Compound child for the Switch, which will collect children in the
    # correct order
    new_compound = Compound([], switch_node.stmt.coord)

    # The last Case/Default node
    last_case = None

    # Goes over the children of the Compound below the Switch, adding them
    # either directly below new_compound or below the last Case as appropriate
    # (for `switch(cond) {}`, block_items would have been None)
    for child in (switch_node.stmt.block_items or []):
        if isinstance(child, (Case, Default)):
            # If it's a Case/Default:
            # 1. Add it to the Compound and mark as "last case"
            # 2. If its immediate child is also a Case or Default, promote it
            #    to a sibling.
            new_compound.block_items.append(child)
            _extract_nested_case(child, new_compound.block_items)
            last_case = new_compound.block_items[-1]
        else:
            # Other statements are added as children to the last case, if it
            # exists.
            if last_case is None:
                new_compound.block_items.append(child)
            else:
                last_case.stmts.append(child)

    switch_node.stmt = new_compound
    return switch_node

def _extract_nested_case(case_node, stmts_list):
    """ Recursively extract consecutive Case statements that are made nested
        by the parser and add them to the stmts_list.
    """
    if isinstance(case_node.stmts[0], (Case, Default)):
        stmts_list.append(case_node.stmts.pop())
        _extract_nested_case(stmts_list[-1], stmts_list)

def fix_atomic_specifiers(decl):
    """ Atomic specifiers like _Atomic(type) are unusually structured,
        conferring a qualifier upon the contained type.

        This function fixes a decl with atomic specifiers to have a sane AST
        structure, by removing spurious Typename->TypeDecl pairs and attaching
        the _Atomic qualifier in the right place.
    """
    # There can be multiple levels of _Atomic in a decl; fix them until a
    # fixed point is reached.
    while True:
        decl, found = _fix_atomic_specifiers_once(decl)
        if not found:
            break

    # Make sure to add an _Atomic qual on the topmost decl if needed. Also
    # restore the declname on the innermost TypeDecl (it gets placed in the
    # wrong place during construction).
    typ = decl
    while not isinstance(typ, TypeDecl):
        try:
            typ = typ.type
        except AttributeError:
            return decl
    if '_Atomic' in typ.quals and '_Atomic' not in decl.quals:
        decl.quals.append('_Atomic')
    if typ.declname is None:
        typ.declname = decl.name

    return decl

def _fix_atomic_specifiers_once(decl):
    """ Performs one 'fix' round of atomic specifiers.
        Returns (modified_decl, found) where found is True iff a fix was made.
    """
    parent = decl
    grandparent = None
    node = decl.type
    while node is not None:
        if isinstance(node, Typename) and '_Atomic' in node.quals:
            break
        try:
            grandparent = parent
            parent = node
            node = node.type
        except AttributeError:
            # If we've reached a node without a `type` field, it means we won't
            # find what we're looking for at this point; give up the search
            # and return the original decl unmodified.
            return decl, False

    assert isinstance(parent, TypeDecl)
    grandparent.type = node.type
    if '_Atomic' not in node.type.quals:
        node.type.quals.append('_Atomic')
    return decl, True
